package data.missions;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.campaign.CampaignFleetAPI;
import com.fs.starfarer.api.campaign.FactionAPI;
import com.fs.starfarer.api.fleet.FleetGoal;
import com.fs.starfarer.api.fleet.FleetMemberAPI;
import com.fs.starfarer.api.impl.campaign.fleets.FleetFactory;
import com.fs.starfarer.api.impl.campaign.ids.Factions;
import com.fs.starfarer.api.mission.FleetSide;
import com.fs.starfarer.api.mission.MissionDefinitionAPI;
import com.fs.starfarer.api.mission.MissionDefinitionPlugin;
import com.fs.starfarer.api.util.Misc;
import org.lazywizard.lazylib.MathUtils;
import org.lazywizard.lazylib.VectorUtils;
import org.lwjgl.util.vector.Vector2f;

/**
 *  Random mission class
 *  Adapted from Exi's one (in turn modified from ICE) with permission
 */

public class BaseRandomSpireMissionDefinition implements MissionDefinitionPlugin
{
    // Types of objectives that may be randomly used
    private static final String[] OBJECTIVE_TYPES =
    {
        "sensor_array", "nav_buoy", "comm_relay"
    };
    protected static final Map<String, Float> QUALITY_FACTORS = new HashMap<>(16);
    protected static final float QUALITY_FACTOR_DEFAULT = 0.7f;
    protected FactionAPI enemy;
    protected boolean flagshipChosen = false;
    protected FactionAPI player;
    protected final Random rand = new Random();

    protected boolean predefinedFleetPlayer = false;
    protected boolean predefinedFleetEnemy = false;
    
    static
    {
        QUALITY_FACTORS.put("spire", 0.8f);
        QUALITY_FACTORS.put("darkspire", 0.8f);
        QUALITY_FACTORS.put(Factions.DIKTAT, 0.5f);
        QUALITY_FACTORS.put(Factions.HEGEMONY, 0.5f);
        QUALITY_FACTORS.put(Factions.INDEPENDENT, 0.5f);
        QUALITY_FACTORS.put(Factions.KOL, 0.5f);
        QUALITY_FACTORS.put(Factions.LIONS_GUARD, 0.75f);
        QUALITY_FACTORS.put(Factions.LUDDIC_CHURCH, 0.25f);
        QUALITY_FACTORS.put(Factions.LUDDIC_PATH, 0f);
        QUALITY_FACTORS.put(Factions.PIRATES, 0f);
        QUALITY_FACTORS.put(Factions.TRITACHYON, 0.85f);
        QUALITY_FACTORS.put("exigency", 1f);
        QUALITY_FACTORS.put("exipirated", 0.6f);
        QUALITY_FACTORS.put("tiandong", 0.55f);
        QUALITY_FACTORS.put("sun_ice", 0.85f);
        QUALITY_FACTORS.put("sun_ici", 0.75f);
    }

    @Override
    public void defineMission(MissionDefinitionAPI api)
    {
        if (player == null || enemy == null)
        {
            chooseFactions(null, null);
        }

        api.initFleet(FleetSide.PLAYER, "SIS", FleetGoal.ATTACK, false, 5);
        api.initFleet(FleetSide.ENEMY, "", FleetGoal.ATTACK, true, 5);

        api.setFleetTagline(FleetSide.PLAYER, Misc.ucFirst(player.getDisplayName()) + " forces");
        api.setFleetTagline(FleetSide.ENEMY, Misc.ucFirst(enemy.getDisplayName()) + " forces");

        // Fleet size randomization
        int size = 30 + (int) ((float) Math.random() * 170);
        float difficulty = 0.7f + rand.nextFloat() * 0.3f;

        // Actual fleet generation call
        String playerFP = "?";
        if (!predefinedFleetPlayer)
        {    
            float qf = QUALITY_FACTOR_DEFAULT;
            if (QUALITY_FACTORS.containsKey(player.getId()))
                qf = QUALITY_FACTORS.get(player.getId());
            playerFP = "" + generateFleet(player, api, FleetSide.PLAYER, (int) (size * difficulty), qf);
        }
        String enemyFP = "?";
        if (!predefinedFleetEnemy)
        {
            float qf = 0.7f;
            if (QUALITY_FACTORS.containsKey(enemy.getId()))
                qf = QUALITY_FACTORS.get(enemy.getId());
            enemyFP = "" + generateFleet(enemy, api, FleetSide.ENEMY, size, qf);
        }

        // Set up the map
        float width = 10000f + 10000f * (size / 100);
        float height = 10000f + 10000f * (size / 100);
        api.initMap(-width / 2f, width / 2f, -height / 2f, height / 2f);

        float minX = -width / 2;
        float minY = -height / 2;

        int objectiveCount = (int) Math.floor(size / 35f);

        while (objectiveCount > 0)
        {
            String type = OBJECTIVE_TYPES[rand.nextInt(3)];

            if (objectiveCount == 1)
            {
                api.addObjective(0, 0, type);
                objectiveCount -= 1;
            }
            else
            {
                float theta = (float) (Math.random() * Math.PI);
                double radius = Math.min(width, height);
                radius = radius * 0.1 + radius * 0.3 * Math.random();
                int x = (int) (Math.cos(theta) * radius);
                int y = (int) (Math.sin(theta) * radius);
                api.addObjective(x, -y, type);
                api.addObjective(-x, y, type);
                objectiveCount -= 2;
            }
        }

        // If the player faction or enemy faction is the spire, there is a 50% chance of planets being added
        if (Math.random() < 0.5 && (player.getId().contentEquals("spire") || enemy.getId().contentEquals("spire")))
        {

            // The possible planet backgrounds
            /*
            String[] planets =
            {
                "exigency_planetoid", "exigency_black_hole"
            };
            String planet = planets[(int) (Math.random() * planets.length)];

            float radius;

            // If the planet is the 'planetoid', set the radius to...
            if (planet.contentEquals("exigency_planetoid"))
            {
                radius = 60f;

                // Spawn Jump Point
                float angle = (float) Math.random() * 360f;
                Vector2f pos = new Vector2f(600f, 0f);
                VectorUtils.rotate(pos, angle, pos);
                api.addPlanet(pos.x, pos.y, 20f, "wormholeUnder", 0f, true);
                api.addPlanet(pos.x, pos.y, 20f, "wormholeA", 0f, true);
                api.addPlanet(pos.x, pos.y, 20f, "wormholeB", 0f, true);
                api.addPlanet(pos.x, pos.y, 20f, "wormholeC", 0f, true);

                // Otherwise it is the blackhole, set it to...
            }
            else
            {
                radius = 300f;
            }
            api.addPlanet(0, 0, radius, planet, 200f, true);
            */

            // Something weird happening with the descriptor setting for the background not working, force it via java
            //api.setBackgroundSpriteName("graphics/AIWar/backgrounds/gemstone.jpg");
        }

        // Show the factions versus and their FP
        api.addBriefingItem(Misc.ucFirst(player.getDisplayName()) + "  (" + playerFP + ")   vs.  " + Misc.ucFirst(enemy.getDisplayName()) + "  (" + enemyFP + ")");

        // Chance of generating a nebula
        float nebulaChance = MathUtils.getRandomNumberInRange(0, 100);

        // So basically half the time (if less than 50 out of 100)
        if (nebulaChance < 50)
        {
            // Do regular nebula generation
            float nebulaCount = 10 + (float) Math.random() * 30;
            float nebulaSize = (float) Math.random();

            for (int i = 0; i < nebulaCount; i++)
            {
                float x = (float) Math.random() * width - width / 2;
                float y = (float) Math.random() * height - height / 2;
                float nebulaRadius = (400f + (float) Math.random() * 1600f) * nebulaSize;
                api.addNebula(x, y, nebulaRadius);
            }

            api.addBriefingItem("Nebulosity:  " + (int) (((nebulaCount * nebulaSize) / 40f) * 100) + "%");

        }
        else
        {
            // Mention that there is no nebula, this line could be commented out if you don't want this line item added
            api.addBriefingItem("Nebulosity: N/A");
        }

        // Asteroid generation random chance
        float asteroidChance = MathUtils.getRandomNumberInRange(0, 100);

        // If chance is less than 50
        if (asteroidChance < 50)
        {

            // Do regular asteroid generation
            // Minimum asteroid speed between 15 and 50
            int minAsteroidSpeed = MathUtils.getRandomNumberInRange(15, 50);

            // Asteroid count
            int asteroidCount = size + (int) (size * 4 * Math.pow(Math.random(), 2));

            // Add the asteroid field
            api.addAsteroidField(
                    minX + width * 0.5f, // X
                    minY + height * 0.5f, // Y
                    rand.nextInt(90) - 45 + (rand.nextInt() % 2) * 180, // Angle
                    100 + (int) (Math.random() * height / 2), // Width
                    minAsteroidSpeed, // Min speed
                    minAsteroidSpeed * 1.1f, // Max speed
                    asteroidCount); // Count

            api.addBriefingItem("Asteroid Density:  " + (int) ((asteroidCount / 1000f) * 100) + "%");
            api.addBriefingItem("Asteroid Speed:  " + minAsteroidSpeed);
        }
        else
        {
            // If not asteroid field, specify as N/A, you can comment this out
            api.addBriefingItem("Asteroid Density: N/A");
            api.addBriefingItem("Asteroid Speed: N/A");
        }

    }

    protected void chooseFactions(String playerFactionId, String enemyFactionId)
    {
        player = Global.getSector().getFaction(playerFactionId);
        enemy = Global.getSector().getFaction(enemyFactionId);

        List<FactionAPI> acceptableFactions = new ArrayList<>();

        // Instead of adding *all* factions, we are adding the vanilla factions only
        /*
        acceptableFactions.add(Global.getSector().getFaction(Factions.DIKTAT));
        acceptableFactions.add(Global.getSector().getFaction(Factions.HEGEMONY));
        acceptableFactions.add(Global.getSector().getFaction(Factions.INDEPENDENT));
        acceptableFactions.add(Global.getSector().getFaction(Factions.KOL));
        acceptableFactions.add(Global.getSector().getFaction(Factions.LIONS_GUARD));
        acceptableFactions.add(Global.getSector().getFaction(Factions.LUDDIC_CHURCH));
        acceptableFactions.add(Global.getSector().getFaction(Factions.LUDDIC_PATH));
        acceptableFactions.add(Global.getSector().getFaction(Factions.PIRATES));
        acceptableFactions.add(Global.getSector().getFaction(Factions.TRITACHYON));
        */
        for (FactionAPI faction : Global.getSector().getAllFactions())
        {
            String id = faction.getId();
            if (faction.isNeutralFaction()) continue;
            if (faction.isPlayerFaction()) continue;
            if (id.equals("player_npc")) continue;
            if (id.equals("merc_hostile")) continue;
            if (id.equals("famous_bounty")) continue;
            if (!id.equals("spire") && !id.equals("darkspire")) 
                acceptableFactions.add(faction);
        }

        /*
         If the player fleetside is not null, and the player faction equals spire
         add the darkspire as a possible faction

         else if the player fleetside is not null and the player faction ID is darkspire
         add the spire as a possible faction
         */
        if (null != player && playerFactionId.contentEquals("spire"))
        {
            acceptableFactions.add(Global.getSector().getFaction("darkspire"));
        }
        else if (null != player && playerFactionId.contentEquals("darkspire"))
        {
            acceptableFactions.add(Global.getSector().getFaction("spire"));
        }

        // Similar if else statement as above except is for the enemy faction ID
        if (null != enemy && enemyFactionId.contentEquals("spire"))
        {
            acceptableFactions.add(Global.getSector().getFaction("darkspire"));
        }
        else if (null != enemy && enemyFactionId.contentEquals("darkspire"))
        {
            acceptableFactions.add(Global.getSector().getFaction("spire"));
        }

        // If the player faction is not null, and is not specified a parameter in the input, choose from one of the acceptable factions
        player = player != null ? player : acceptableFactions.get(rand.nextInt(acceptableFactions.size()));

        acceptableFactions.remove(player);

        enemy = enemy != null ? enemy : acceptableFactions.get(rand.nextInt(acceptableFactions.size()));
    }
    
    protected void usePredefinedFleet(boolean player, boolean enemy)
    {
        predefinedFleetPlayer = player;
        predefinedFleetEnemy = enemy;
    }

    // Generate a fleet from the campaign fleet generator
    int generateFleet(FactionAPI faction, MissionDefinitionAPI api, FleetSide side, int minFP, float qf)
    {
        CampaignFleetAPI fleet = FleetFactory.createGenericFleet(faction.getId(), side.toString(), qf, minFP);
        fleet.getFleetData().sort();

        for (FleetMemberAPI m : fleet.getFleetData().getMembersListCopy())
        {
            String variant = m.isFighterWing() ? m.getSpecId() : m.getVariant().getHullVariantId();
            api.addToFleet(side, variant, m.getType(), m.getShipName(), (!m.isFighterWing() && !flagshipChosen), m.getCrewXPLevel());

            if (!m.isFighterWing() && !flagshipChosen)
            {
                flagshipChosen = true;
            }
        }

        return fleet.getFleetPoints();
    }
}